@@ -1,5 +1,5 @@
 /* Basic dependency engine for GNU Make.
-Copyright (C) 1988-2022 Free Software Foundation, Inc.
+Copyright (C) 1988-2023 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify it under the
@@ -12,7 +12,7 @@ WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
-this program.  If not, see <http://www.gnu.org/licenses/>.  */
+this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #include "makeint.h"
 #include "filedef.h"
@@ -34,10 +34,11 @@ this program.  If not, see <http://www.gnu.org/licenses/>.  */
 #include <starlet.h>
 #endif
 #ifdef WINDOWS32
+#include <windows.h>
 #include <io.h>
 #include <sys/stat.h>
 #if defined(_MSC_VER) && _MSC_VER > 1200
-/* VC7 or later supprots _stat64 to access 64-bit file size. */
+/* VC7 or later supports _stat64 to access 64-bit file size. */
 #define STAT _stat64
 #else
 #define STAT stat
@@ -78,6 +79,25 @@ static FILE_TIMESTAMP name_mtime (const char *name);
 static const char *library_search (const char *lib, FILE_TIMESTAMP *mtime_ptr);
 
 
+static void
+check_also_make (const struct file *file)
+{
+  struct dep *ad;
+  FILE_TIMESTAMP mtime = file->last_mtime;
+
+  if (mtime == UNKNOWN_MTIME)
+    mtime = name_mtime (file->name);
+
+  /* If we updated the file, check its also-make files.  */
+
+  if (is_ordinary_mtime (mtime) && mtime > file->mtime_before_update)
+    for (ad = file->also_make; ad; ad = ad->next)
+      if (ad->file->last_mtime == NONEXISTENT_MTIME)
+        OS (error, file->cmds ? &file->cmds->fileinfo : NILF,
+            _("warning: pattern recipe did not update peer target '%s'."),
+            ad->file->name);
+}
+
 /* Remake all the goals in the 'struct dep' chain GOALS.  Return update_status
    representing the totality of the status of the goals.
 
@@ -89,6 +109,7 @@ static const char *library_search (const char *lib, FILE_TIMESTAMP *mtime_ptr);
 enum update_status
 update_goal_chain (struct goaldep *goaldeps)
 {
+  unsigned long last_cmd_count = 0;
   int t = touch_flag, q = question_flag, n = just_print_flag;
   enum update_status status = us_none;
 
@@ -114,9 +135,12 @@ update_goal_chain (struct goaldep *goaldeps)
 
       start_waiting_jobs ();
 
-      /* Wait for a child to die.  */
+      /* Check for exited children.  If no children have finished since the
+         last time we looked, then block until one exits.  If some have
+         exited don't block, so we can possibly do more work.  */
 
-      reap_children (1, 0);
+      reap_children (last_cmd_count == command_count, 0);
+      last_cmd_count = command_count;
 
       lastgoal = 0;
       gu = goals;
@@ -416,7 +440,7 @@ complain (struct file *file)
 }
 
 /* Consider a single 'struct file' and update it as appropriate.
-   Return 0 on success, or non-0 on failure.  */
+   Return an update_status value; use us_success if we aren't sure yet.  */
 
 static enum update_status
 update_file_1 (struct file *file, unsigned int depth)
@@ -449,7 +473,7 @@ update_file_1 (struct file *file, unsigned int depth)
         }
 
       DBF (DB_VERBOSE, _("File '%s' was considered already.\n"));
-      return 0;
+      return us_success;
     }
 
   switch (file->command_state)
@@ -459,7 +483,7 @@ update_file_1 (struct file *file, unsigned int depth)
       break;
     case cs_running:
       DBF (DB_VERBOSE, _("Still updating file '%s'.\n"));
-      return 0;
+      return us_success;
     case cs_finished:
       DBF (DB_VERBOSE, _("Finished updating file '%s'.\n"));
       return file->update_status;
@@ -471,8 +495,6 @@ update_file_1 (struct file *file, unsigned int depth)
      fail. */
   file->no_diag = file->dontcare;
 
-  ++depth;
-
   /* Notice recursive update of the same file.  */
   start_updating (file);
 
@@ -480,6 +502,9 @@ update_file_1 (struct file *file, unsigned int depth)
      remember this one to turn off updating.  */
   ofile = file;
 
+  /* Increase the depth for reporting how we build the file.  */
+  ++depth;
+
   /* Looking at the file's modtime beforehand allows the possibility
      that its name may be changed by a VPATH search, and thus it may
      not need an implicit rule.  If this were not done, the file
@@ -491,8 +516,7 @@ update_file_1 (struct file *file, unsigned int depth)
   noexist = this_mtime == NONEXISTENT_MTIME;
   if (noexist)
     DBF (DB_BASIC, _("File '%s' does not exist.\n"));
-  else if (ORDINARY_MTIME_MIN <= this_mtime && this_mtime <= ORDINARY_MTIME_MAX
-           && file->low_resolution_time)
+  else if (is_ordinary_mtime (this_mtime) && file->low_resolution_time)
     {
       /* Avoid spurious rebuilds due to low resolution time stamps.  */
       int ns = FILE_TIMESTAMP_NS (this_mtime);
@@ -503,6 +527,28 @@ update_file_1 (struct file *file, unsigned int depth)
       this_mtime += FILE_TIMESTAMPS_PER_S - 1 - ns;
     }
 
+  /* If any also_make target doesn't exist, we must remake this one too.
+     If they do exist choose the oldest mtime so they will rebuild.  */
+
+  for (ad = file->also_make; ad && !noexist; ad = ad->next)
+    {
+      struct file *adfile = ad->file;
+      FILE_TIMESTAMP fmtime = file_mtime (adfile);
+
+      noexist = fmtime == NONEXISTENT_MTIME;
+      if (noexist)
+        {
+          check_renamed (adfile);
+          DBS (DB_BASIC,
+               (_("Grouped target peer '%s' of file '%s' does not exist.\n"),
+                adfile->name, file->name));
+        }
+      else if (fmtime < this_mtime)
+        this_mtime = fmtime;
+    }
+
+  must_make = noexist;
+
   /* If file was specified as a target with no commands, come up with some
      default commands.  This may also add more also_make files.  */
 
@@ -518,26 +564,6 @@ update_file_1 (struct file *file, unsigned int depth)
       file->cmds = default_file->cmds;
     }
 
-  /* If any also_make target doesn't exist, we must remake this one too.
-     If they do exist choose the oldest mtime so they will rebuild.  */
-
-  for (ad = file->also_make; ad && !noexist; ad = ad->next)
-    {
-      struct file *adfile = ad->file;
-      FILE_TIMESTAMP fmtime = file_mtime (adfile);
-
-      check_renamed (adfile);
-      noexist = fmtime == NONEXISTENT_MTIME;
-      if (noexist)
-        DBS (DB_BASIC,
-             (_("Grouped target peer '%s' of file '%s' does not exist.\n"),
-              adfile->name, file->name));
-      else if (fmtime < this_mtime)
-        this_mtime = fmtime;
-    }
-
-  must_make = noexist;
-
   /* Update all non-intermediate files we depend on, if necessary, and see
      whether any of them is more recent than this file.  We need to walk our
      deps, AND the deps of any also_make targets to ensure everything happens
@@ -713,16 +739,18 @@ update_file_1 (struct file *file, unsigned int depth)
   finish_updating (file);
   finish_updating (ofile);
 
-  DBF (DB_VERBOSE, _("Finished prerequisites of target file '%s'.\n"));
+  /* We've decided what we need to do to build the file.  */
+  --depth;
 
   if (running)
     {
       set_command_state (file, cs_deps_running);
-      --depth;
       DBF (DB_VERBOSE, _("The prerequisites of '%s' are being made.\n"));
-      return 0;
+      return us_success;
     }
 
+  DBF (DB_VERBOSE, _("Finished prerequisites of target file '%s'.\n"));
+
   /* If any dependency failed, give up now.  */
 
   if (dep_status)
@@ -731,8 +759,6 @@ update_file_1 (struct file *file, unsigned int depth)
       file->update_status = dep_status == us_none ? us_failed : dep_status;
       notice_finished_file (file);
 
-      --depth;
-
       DBF (DB_VERBOSE, _("Giving up on target file '%s'.\n"));
 
       if (depth == 0 && keep_going_flag
@@ -807,16 +833,13 @@ update_file_1 (struct file *file, unsigned int depth)
 
           if (fmt)
             {
-              print_spaces (depth);
+              print_spaces (depth+1);
               printf (fmt, dep_name (d), file->name);
               fflush (stdout);
             }
         }
     }
 
-  /* Here depth returns to the value it had when we were called.  */
-  depth--;
-
   if (file->double_colon && file->deps == 0)
     {
       must_make = 1;
@@ -850,7 +873,8 @@ update_file_1 (struct file *file, unsigned int depth)
 
       /* Since make has not created this file, make should not remove it,
          even if the file is intermediate. */
-      file->secondary = 1;
+      if (!file->notintermediate && no_intermediates == 0)
+        file->secondary = 1;
 
       notice_finished_file (file);
 
@@ -864,7 +888,7 @@ update_file_1 (struct file *file, unsigned int depth)
           file = file->prev;
         }
 
-      return 0;
+      return us_success;
     }
 
   DBF (DB_BASIC, _("Must remake target '%s'.\n"));
@@ -883,7 +907,7 @@ update_file_1 (struct file *file, unsigned int depth)
   if (file->command_state != cs_finished)
     {
       DBF (DB_VERBOSE, _("Recipe of '%s' is being run.\n"));
-      return 0;
+      return us_success;
     }
 
   switch (file->update_status)
@@ -936,7 +960,7 @@ notice_finished_file (struct file *file)
              we don't want to do the touching.  */
           unsigned int i;
           for (i = 0; i < file->cmds->ncommand_lines; ++i)
-            if (!(file->cmds->lines_flags[i] & COMMANDS_RECURSE))
+            if (NONE_SET (file->cmds->lines_flags[i], COMMANDS_RECURSE))
               goto have_nonrecursing;
         }
       else
@@ -977,7 +1001,7 @@ notice_finished_file (struct file *file)
       if ((question_flag || just_print_flag || touch_flag) && file->cmds)
         {
           for (i = file->cmds->ncommand_lines; i > 0; --i)
-            if (! (file->cmds->lines_flags[i-1] & COMMANDS_RECURSE))
+            if (NONE_SET (file->cmds->lines_flags[i-1], COMMANDS_RECURSE))
               break;
         }
 
@@ -1018,23 +1042,30 @@ notice_finished_file (struct file *file)
     }
 
   if (ran && file->update_status != us_none)
-    /* We actually tried to update FILE, which has
-       updated its also_make's as well (if it worked).
-       If it didn't work, it wouldn't work again for them.
-       So mark them as updated with the same status.  */
-    for (d = file->also_make; d != 0; d = d->next)
-      {
-        d->file->command_state = cs_finished;
-        d->file->updated = 1;
-        d->file->update_status = file->update_status;
+    {
+      /* We actually tried to update FILE, which has
+         updated its also_make's as well (if it worked).
+         If it didn't work, it wouldn't work again for them.
+         So mark them as updated with the same status.  */
+      for (d = file->also_make; d != 0; d = d->next)
+        {
+          d->file->command_state = cs_finished;
+          d->file->updated = 1;
+          d->file->update_status = file->update_status;
 
-        if (ran && !d->file->phony)
-          /* Fetch the new modification time.
-             We do this instead of just invalidating the cached time
-             so that a vpath_search can happen.  Otherwise, it would
-             never be done because the target is already updated.  */
-          f_mtime (d->file, 0);
-      }
+          if (ran && !d->file->phony)
+            /* Fetch the new modification time.
+               We do this instead of just invalidating the cached time
+               so that a vpath_search can happen.  Otherwise, it would
+               never be done because the target is already updated.  */
+            f_mtime (d->file, 0);
+        }
+
+      /* If the target was created by an implicit rule, and it was updated,
+         warn about any of its also_make targets that don't exist.  */
+      if (file->tried_implicit && file->also_make)
+        check_also_make (file);
+    }
   else if (file->update_status == us_none)
     /* Nothing was done for FILE, but it needed nothing done.
        So mark it now as "succeeded".  */
@@ -1055,7 +1086,6 @@ check_dep (struct file *file, unsigned int depth,
   struct dep *d;
   enum update_status dep_status = us_success;
 
-  ++depth;
   start_updating (file);
 
   /* We might change file if we find a different one via vpath;
@@ -1153,7 +1183,7 @@ check_dep (struct file *file, unsigned int depth,
 
               d->file->parent = file;
               maybe_make = *must_make_ptr;
-              new = check_dep (d->file, depth, this_mtime, &maybe_make);
+              new = check_dep (d->file, depth+1, this_mtime, &maybe_make);
               if (new > dep_status)
                 dep_status = new;
 
@@ -1312,7 +1342,7 @@ f_mtime (struct file *file, int search)
   if (ar_name (file->name))
     {
       /* This file is an archive-member reference.  */
-
+      FILE_TIMESTAMP memmtime;
       char *arname, *memname;
       struct file *arfile;
       time_t member_date;
@@ -1320,6 +1350,9 @@ f_mtime (struct file *file, int search)
       /* Find the archive's name.  */
       ar_parse_name (file->name, &arname, &memname);
 
+      /* Find the mtime of the member file (it might not exist).  */
+      memmtime = name_mtime (memname);
+
       /* Find the modification time of the archive itself.
          Also allow for its name to be changed via VPATH search.  */
       arfile = lookup_file (arname);
@@ -1363,9 +1396,16 @@ f_mtime (struct file *file, int search)
         return NONEXISTENT_MTIME;
 
       member_date = ar_member_date (file->hname);
-      mtime = (member_date == (time_t) -1
-               ? NONEXISTENT_MTIME
-               : file_timestamp_cons (file->hname, member_date, 0));
+
+      if (member_date == (time_t) -1
+          || (memmtime != NONEXISTENT_MTIME
+              && (time_t) FILE_TIMESTAMP_S (memmtime) > member_date))
+        /* If the member file exists and is newer than the member in the
+           archive, pretend it's nonexistent.  This means the member file was
+           updated but not added to the archive yet.  */
+        mtime = NONEXISTENT_MTIME;
+      else
+        mtime = file_timestamp_cons (file->hname, member_date, 0);
     }
   else
 #endif
@@ -1438,14 +1478,6 @@ f_mtime (struct file *file, int search)
       FILE_TIMESTAMP adjustment = FAT_ADJ_OFFSET << FILE_TIMESTAMP_LO_BITS;
       if (ORDINARY_MTIME_MIN + adjustment <= adjusted_mtime)
         adjusted_mtime -= adjustment;
-#elif defined(__EMX__)
-      /* FAT filesystems round time to the nearest even second!
-         Allow for any file (NTFS or FAT) to perhaps suffer from this
-         brain damage.  */
-      FILE_TIMESTAMP adjustment = (((FILE_TIMESTAMP_S (adjusted_mtime) & 1) == 0
-                     && FILE_TIMESTAMP_NS (adjusted_mtime) == 0)
-                    ? (FILE_TIMESTAMP) 1 << FILE_TIMESTAMP_LO_BITS
-                    : 0);
 #endif
 
       /* If the file's time appears to be in the future, update our
@@ -1524,7 +1556,7 @@ name_mtime (const char *name)
 
 #if defined(WINDOWS32)
   {
-    char tem[MAXPATHLEN], *tstart, *tend;
+    char tem[MAX_PATH+1], *tstart, *tend;
     const char *p = name + strlen (name);
 
     /* Remove any trailing slashes and "."/"..".  MS-Windows stat
@@ -1542,7 +1574,7 @@ name_mtime (const char *name)
           tend--;
         if (*tend == '.' && tend > tstart)
           tend--;
-        for ( ; tend > tstart && (*tend == '/' || *tend == '\\'); tend--)
+        for ( ; tend > tstart && ISDIRSEP (*tend); tend--)
           *tend = '\0';
       }
     else
