@@ -1,5 +1,5 @@
 /* Internals of variables for GNU Make.
-Copyright (C) 1988-2022 Free Software Foundation, Inc.
+Copyright (C) 1988-2023 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify it under the
@@ -12,7 +12,7 @@ WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
-this program.  If not, see <http://www.gnu.org/licenses/>.  */
+this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #include "makeint.h"
 
@@ -472,7 +472,7 @@ lookup_variable (const char *name, size_t length)
       const struct variable_set *set = setlist->set;
       struct variable *v;
 
-      v = (struct variable *) hash_find_item ((struct hash_table *) &set->table, &var_key);
+      v = hash_find_item ((struct hash_table *) &set->table, &var_key);
       if (v && (!is_parent || !v->private_var))
         return v->special ? lookup_special_var (v) : v;
 
@@ -538,6 +538,29 @@ lookup_variable (const char *name, size_t length)
 
   return 0;
 }
+/* Lookup a variable whose name is a string starting at NAME
+   and with LENGTH chars.  NAME need not be null-terminated.
+   Returns address of the 'struct variable' containing all info
+   on the variable, or nil if no such variable is defined.  */
+
+struct variable *
+lookup_variable_for_file (const char *name, size_t length, struct file *file)
+{
+  struct variable *var;
+  struct variable_set_list *savev;
+
+  if (file == NULL)
+    return lookup_variable (name, length);
+
+  savev = current_variable_set_list;
+  current_variable_set_list = file->variables;
+
+  var = lookup_variable (name, length);
+
+  current_variable_set_list = savev;
+
+  return var;
+}
 
 /* Lookup a variable whose name is a string starting at NAME
    and with LENGTH chars in set SET.  NAME need not be null-terminated.
@@ -553,7 +576,7 @@ lookup_variable_in_set (const char *name, size_t length,
   var_key.name = (char *) name;
   var_key.length = (unsigned int) length;
 
-  return (struct variable *) hash_find_item ((struct hash_table *) &set->table, &var_key);
+  return hash_find_item ((struct hash_table *) &set->table, &var_key);
 }
 
 /* Initialize FILE's variable set list.  If FILE already has a variable set
@@ -905,7 +928,7 @@ define_automatic_variables (void)
     if (!replace || !*replace->value)
       replace = lookup_variable ("OS2_SHELL", 9);
 # else
-#   warning NO_CMD_DEFAULT: GNU make will not use CMD.EXE as default shell
+#   warning NO_CMD_DEFAULT: GNU Make will not use CMD.EXE as default shell
 # endif
 
     if (replace && *replace->value)
@@ -1065,7 +1088,8 @@ target_environment (struct file *file, int recursive)
   for (s = set_list; s != 0; s = s->next)
     {
       struct variable_set *set = s->set;
-      int isglobal = set == &global_variable_set;
+      const int islocal = s == set_list;
+      const int isglobal = set == &global_variable_set;
 
       v_slot = (struct variable **) set->table.ht_vec;
       v_end = v_slot + set->table.ht_size;
@@ -1075,11 +1099,17 @@ target_environment (struct file *file, int recursive)
             struct variable **evslot;
             struct variable *v = *v_slot;
 
+            if (!islocal && v->private_var)
+              continue;
+
             evslot = (struct variable **) hash_find_slot (&table, v);
 
             if (HASH_VACANT (*evslot))
               {
-                /* If we're not global, or we are and should export, add it.  */
+                /* We'll always add target-specific variables, since we may
+                   discover that they should be exported later: we'll check
+                   again below.  For global variables only add them if they're
+                   exportable.  */
                 if (!isglobal || should_export (v))
                   hash_insert_at (&table, v, evslot);
               }
@@ -1107,8 +1137,9 @@ target_environment (struct file *file, int recursive)
 
         /* If V is recursively expanded and didn't come from the environment,
            expand its value.  If it came from the environment, it should
-           go back into the environment unchanged.  */
-        if (v->recursive && v->origin != o_env && v->origin != o_env_override)
+           go back into the environment unchanged... except MAKEFLAGS.  */
+        if (v->recursive && ((v->origin != o_env && v->origin != o_env_override)
+                             || streq (v->name, MAKEFLAGS_NAME)))
           value = cp = recursively_expand_for_file (v, file);
 
         /* If this is the SHELL variable remember we already added it.  */
@@ -1215,17 +1246,18 @@ target_environment (struct file *file, int recursive)
 }
 
 static struct variable *
-set_special_var (struct variable *var)
+set_special_var (struct variable *var, enum variable_origin origin)
 {
-  if (streq (var->name, RECIPEPREFIX_NAME))
+  if (streq (var->name, MAKEFLAGS_NAME))
+    reset_makeflags (origin);
+
+  else if (streq (var->name, RECIPEPREFIX_NAME))
     {
       /* The user is resetting the command introduction prefix.  This has to
          happen immediately, so that subsequent rules are interpreted
          properly.  */
       cmd_prefix = var->value[0]=='\0' ? RECIPEPREFIX_DEFAULT : var->value[0];
     }
-  else if (streq (var->name, MAKEFLAGS_NAME))
-    decode_env_switches (STRING_SIZE_TUPLE(MAKEFLAGS_NAME));
 
   return var;
 }
@@ -1261,7 +1293,7 @@ do_variable_definition (const floc *flocp, const char *varname,
                         const char *value, enum variable_origin origin,
                         enum variable_flavor flavor, int target_var)
 {
-  const char *p;
+  const char *newval;
   char *alloc_value = NULL;
   struct variable *v;
   int append = 0;
@@ -1276,7 +1308,7 @@ do_variable_definition (const floc *flocp, const char *varname,
          We have to allocate memory since otherwise it'll clobber the
          variable buffer, and we may still need that if we're looking at a
          target-specific variable.  */
-      p = alloc_value = allocated_variable_expand (value);
+      newval = alloc_value = allocated_variable_expand (value);
       break;
     case f_expand:
       {
@@ -1285,16 +1317,16 @@ do_variable_definition (const floc *flocp, const char *varname,
            tokens to '$$' to resolve to '$' when recursively expanded.  */
         char *t = allocated_variable_expand (value);
         char *np = alloc_value = xmalloc (strlen (t) * 2 + 1);
-        p = t;
-        while (p[0] != '\0')
+        char *op = t;
+        while (op[0] != '\0')
           {
-            if (p[0] == '$')
+            if (op[0] == '$')
               *(np++) = '$';
-            *(np++) = *(p++);
+            *(np++) = *(op++);
           }
         *np = '\0';
-        p = alloc_value;
         free (t);
+        newval = alloc_value;
         break;
       }
     case f_shell:
@@ -1302,9 +1334,10 @@ do_variable_definition (const floc *flocp, const char *varname,
         /* A shell definition "var != value".  Expand value, pass it to
            the shell, and store the result in recursively-expanded var. */
         char *q = allocated_variable_expand (value);
-        p = alloc_value = shell_result (q);
+        alloc_value = shell_result (q);
         free (q);
         flavor = f_recursive;
+        newval = alloc_value;
         break;
       }
     case f_conditional:
@@ -1320,7 +1353,7 @@ do_variable_definition (const floc *flocp, const char *varname,
     case f_recursive:
       /* A recursive variable definition "var = value".
          The value is used verbatim.  */
-      p = value;
+      newval = value;
       break;
     case f_append:
     case f_append_value:
@@ -1345,15 +1378,16 @@ do_variable_definition (const floc *flocp, const char *varname,
           {
             /* There was no old value.
                This becomes a normal recursive definition.  */
-            p = value;
+            newval = value;
             flavor = f_recursive;
           }
         else
           {
             /* Paste the old and new values together in VALUE.  */
 
-            size_t oldlen, vallen;
+            size_t oldlen, vallen, alloclen;
             const char *val;
+            char *cp;
             char *tp = NULL;
 
             val = value;
@@ -1378,18 +1412,25 @@ do_variable_definition (const floc *flocp, const char *varname,
               }
 
             oldlen = strlen (v->value);
-            p = alloc_value = xmalloc (oldlen + 1 + vallen + 1);
+            alloclen = oldlen + 1 + vallen + 1;
+            cp = alloc_value = xmalloc (alloclen);
 
             if (oldlen)
               {
-                memcpy (alloc_value, v->value, oldlen);
-                alloc_value[oldlen] = ' ';
-                ++oldlen;
+                char *s;
+                if (streq (varname, MAKEFLAGS_NAME)
+                    && (s = strstr (v->value, " -- ")))
+                  /* We found a separator in MAKEFLAGS.  Ignore variable
+                     assignments: set_special_var() will reconstruct things.  */
+                  cp = mempcpy (cp, v->value, s - v->value);
+                else
+                  cp = mempcpy (cp, v->value, oldlen);
+                *(cp++) = ' ';
               }
 
-            memcpy (&alloc_value[oldlen], val, vallen + 1);
-
+            memcpy (cp, val, vallen + 1);
             free (tp);
+            newval = alloc_value;
           }
       }
       break;
@@ -1399,6 +1440,8 @@ do_variable_definition (const floc *flocp, const char *varname,
       abort ();
     }
 
+  assert (newval);
+
 #ifdef __MSDOS__
   /* Many Unix Makefiles include a line saying "SHELL=/bin/sh", but
      non-Unix systems don't conform to this default configuration (in
@@ -1440,16 +1483,16 @@ do_variable_definition (const floc *flocp, const char *varname,
           char *fake_env[2];
           size_t pathlen = 0;
 
-          shellbase = strrchr (p, '/');
-          bslash = strrchr (p, '\\');
+          shellbase = strrchr (newval, '/');
+          bslash = strrchr (newval, '\\');
           if (!shellbase || bslash > shellbase)
             shellbase = bslash;
-          if (!shellbase && p[1] == ':')
-            shellbase = p + 1;
+          if (!shellbase && newval[1] == ':')
+            shellbase = newval + 1;
           if (shellbase)
             shellbase++;
           else
-            shellbase = p;
+            shellbase = newval;
 
           /* Search for the basename of the shell (with standard
              executable extensions) along the $PATH.  */
@@ -1490,7 +1533,7 @@ do_variable_definition (const floc *flocp, const char *varname,
          set no_default_sh_exe to indicate sh was found and
          set new value for SHELL variable.  */
 
-      if (find_and_set_default_shell (p))
+      if (find_and_set_default_shell (newval))
         {
           v = define_variable_in_set (varname, strlen (varname), default_shell,
                                       origin, flavor == f_recursive,
@@ -1504,11 +1547,11 @@ do_variable_definition (const floc *flocp, const char *varname,
         {
           char *tp = alloc_value;
 
-          alloc_value = allocated_variable_expand (p);
+          alloc_value = allocated_variable_expand (newval);
 
           if (find_and_set_default_shell (alloc_value))
             {
-              v = define_variable_in_set (varname, strlen (varname), p,
+              v = define_variable_in_set (varname, strlen (varname), newval,
                                           origin, flavor == f_recursive,
                                           (target_var
                                            ? current_variable_set_list->set
@@ -1536,7 +1579,7 @@ do_variable_definition (const floc *flocp, const char *varname,
      invoked in places where we want to define globally visible variables,
      make sure we define this variable in the global set.  */
 
-  v = define_variable_in_set (varname, strlen (varname), p, origin,
+  v = define_variable_in_set (varname, strlen (varname), newval, origin,
                               flavor == f_recursive || flavor == f_expand,
                               (target_var
                                ? current_variable_set_list->set : NULL),
@@ -1546,7 +1589,7 @@ do_variable_definition (const floc *flocp, const char *varname,
 
  done:
   free (alloc_value);
-  return v->special ? set_special_var (v) : v;
+  return v->special ? set_special_var (v, origin) : v;
 }
 
 /* Parse P (a null-terminated string) as a variable definition.
@@ -1770,6 +1813,44 @@ try_variable_definition (const floc *flocp, const char *line,
 
   return vp;
 }
+
+/* These variables are internal to make, and so considered "defined" for the
+   purposes of warn_undefined even if they are not really defined.  */
+
+struct defined_vars
+  {
+    const char *name;
+    size_t len;
+  };
+
+static const struct defined_vars defined_vars[] = {
+  { STRING_SIZE_TUPLE ("MAKECMDGOALS") },
+  { STRING_SIZE_TUPLE ("MAKE_RESTARTS") },
+  { STRING_SIZE_TUPLE ("MAKE_TERMOUT") },
+  { STRING_SIZE_TUPLE ("MAKE_TERMERR") },
+  { STRING_SIZE_TUPLE ("MAKEOVERRIDES") },
+  { STRING_SIZE_TUPLE (".DEFAULT") },
+  { STRING_SIZE_TUPLE ("-*-command-variables-*-") },
+  { STRING_SIZE_TUPLE ("-*-eval-flags-*-") },
+  { STRING_SIZE_TUPLE ("VPATH") },
+  { STRING_SIZE_TUPLE ("GPATH") },
+  { NULL, 0 }
+};
+
+void
+warn_undefined (const char *name, size_t len)
+{
+  if (warn_undefined_variables_flag)
+    {
+      const struct defined_vars *dp;
+      for (dp = defined_vars; dp->name != NULL; ++dp)
+        if (dp->len == len && memcmp (dp->name, name, len) == 0)
+          return;
+
+      error (reading_file, len, _("warning: undefined variable '%.*s'"),
+             (int)len, name);
+    }
+}
 
 /* Print information for variable V, prefixing it with PREFIX.  */
 
@@ -1947,8 +2028,10 @@ sync_Path_environment ()
   if (!path)
     return;
 
-  /* Convert PATH into something WINDOWS32 world can grok.  */
-  convert_Path_to_windows32 (path, ';');
+  /* Convert the value of PATH into something WINDOWS32 world can grok.
+    Note: convert_Path_to_windows32 must see only the value of PATH,
+    and see it from its first character, to do its tricky job.  */
+  convert_Path_to_windows32 (path + CSTRLEN ("PATH="), ';');
 
   environ_path = path;
   putenv (environ_path);
